#!/usr/bin/env python

'''
@author: szhelev

If this code works, it was written by Svetoslav Zhelev. If not then I don't know who wrote it.
'''

from Properties import Properties
from shutil import *
import logging
from xml.dom import minidom, Node
from xml.dom.minidom import Document
import feedparser
import os
import getpass
import socket
import string
from xml.sax._exceptions import SAXParseException
import urllib2
import threading
from urlparse import urlparse, urlunparse
from urllib import urlencode

PROPERTIES_FILE = 'conf/notifier.properties'
LOG_FILE = 'conf/CERNAlerter.log'
FEEDS_FILE = 'conf/feeds.xml'
CERN_ALERTER_DIR_NAME = '.CERNAlerter'

# Specific CERN notifier exception (not used for now) 
class CERNNotifierException(Exception):
    pass

# feed exceptions
class CERNNotifierFeedException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class CERNNotifierFeedAuthException(CERNNotifierFeedException):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class CERNUtils:
    # add get param to URL
    def __parametrizeURL(self, url, **params):
        # don't just add the **params because the url itself
        # might contain CGI variables embedded inside the string.
        url_parsed = list(urlparse(url))
    
        encoded = urlencode(params)
        qs = url_parsed[4]
        if encoded:
            if qs:
                qs += '&' + encoded
            else:
                qs = encoded
        netloc = url_parsed[1]
        if netloc.find('?') > -1:
            url_parsed[1] = url_parsed[1][:netloc.find('?')]
            if qs:
                qs = netloc[netloc.find('?') + 1:] + '&' + qs
            else:
                qs = netloc[netloc.find('?') + 1:]
    
        url_parsed[4] = qs
    
        url = urlunparse(url_parsed)
        return url 
    
    # build url for getting CERN feeds
    def buildCERNFeedURL(self, url):
        configuration = CERNConfiguration.getInstance()
        return self.__parametrizeURL(url, login=configuration.username,
                                                     computername=configuration.computername)
    # get hoostname from URL
    def getHostname(self, url):
        result = urlparse(url)
        return result[1]
        
# This class holds the CERN configuration 
class CERNConfiguration:
    
    # Create a class variable that will hold a reference
    # to the single instance of TestSingleton.
    instance = None
    
    # logger object
    logger = None
    
    # Property names
    P_USERNAME = 'username'
    P_COMPUTERNAME = 'computername'
    
    # Configuration properties
    username = None
    computername = None
    feeds = []

    # Define a helper class that will override the __call___
    # method in order to provide a factory method for the Singleton.
    class SingletonHelper :

        def __call__(self, *args, **kw) :

            # If an instance of CERNConfiguration does not exist,
            # create one and assign it to CERNConfiguration.instance.
            if CERNConfiguration.instance is None :
                object = CERNConfiguration()
                CERNConfiguration.instance = object

            # Return CERNConfiguration.instance, which should contain
            # a reference to the only instance of CERNConfiguration
            # in the system.
            return CERNConfiguration.instance

    # Create a class level method that must be called to
    # get the single instance of CERNConfiguration.
    getInstance = SingletonHelper()
    
    def getLogger(self):
        return self.logger

    # Initialize an instance of the CERNConfiguration class.
    def __init__(self) :

        if not CERNConfiguration.instance == None :
            raise RuntimeError, 'Only one instance of CERNConfiguration is allowed!'
        
        self.logger = logging.getLogger('CERNAlerter')
        hdlr = logging.FileHandler(self.loadFromUsersHomeDir(LOG_FILE))
        formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
        hdlr.setFormatter(formatter)
        self.logger.addHandler(hdlr) 
        self.logger.setLevel(logging.INFO)
        
        self.__loadConfig()


    # load application configuration data
    def __loadConfig(self):
        self.getLogger().debug("Loading configuration.")
        # init data from the property file
        p = Properties()
        p.load(open(self.loadFromUsersHomeDir(PROPERTIES_FILE)))
        
        tmpuser = p.getProperty(CERNConfiguration.P_USERNAME)
        # if there is configured user in properties use it, otherwise use the login user
        if(tmpuser != None and tmpuser != ''):
            self.username = tmpuser
        else:
            self.username = getpass.getuser()
        
        tmpcomputername = p.getProperty(CERNConfiguration.P_COMPUTERNAME)
        # if there is configured computer name in properties use it, otherwise use the computer hostname
        if(tmpcomputername != None and tmpcomputername != ''):
            self.computername = tmpcomputername
        else:
            # get the hostname
            self.computername = socket.gethostname()
            
        # init feeds
        doc = minidom.parse(self.loadFromUsersHomeDir(FEEDS_FILE))
        rootNode = doc.documentElement
        self.__readFeedConfig(rootNode, True, None)
    
    # return config file from user home dir. if it is not there then 
    # just copy the default one there     
    def loadFromUsersHomeDir(self, file):
        user = getpass.getuser()
        homeDir = os.path.expanduser('~' + user)
        alerterDir = homeDir + "/" + CERN_ALERTER_DIR_NAME
        if not os.path.exists(alerterDir):
            os.mkdir(alerterDir)
        confDir = alerterDir + "/conf"
        if not os.path.exists(confDir):
            os.mkdir(confDir)
        fullPath = alerterDir + "/" + file
        if not os.path.isfile(fullPath):
            if os.path.isfile(file):
                copyfile(file, fullPath)
            
        return fullPath
    
    # read feeds from feed configuration file    
    def __readFeedConfig(self, parent, isRoot, currentFeed):
        for node in parent.childNodes:
            if(isRoot):
                tmpFeed = CERNFeedConfiguration()
            if node.nodeType == Node.ELEMENT_NODE:
                content = []
                for child in node.childNodes:
                    if child.nodeType == Node.TEXT_NODE:
                        content.append(child.nodeValue)
                if content:
                    strContent = string.join(content)
                if(not isRoot):
                    if(node.nodeName == CERNFeedConfiguration.N_URL):
                        currentFeed.url = strContent.strip()
                    if(node.nodeName == CERNFeedConfiguration.N_NAME):
                        currentFeed.name = strContent.strip()
                    if(node.nodeName == CERNFeedConfiguration.N_MANDATORY):
                        currentFeed.mandatory = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_CERN_FEED):
                        currentFeed.CERNFeed = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_DISABLED):
                        currentFeed.disabled = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_BALOON_FOR_INFO_MSG):
                        currentFeed.balloonForInfoMsg = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_BALLOON_FOR_IMPORTANT_MSG):
                        currentFeed.balloonForImportantMsg = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_BALLOON_FOR_ALERT_MSG):
                        currentFeed.balloonForAlertMsg = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_POPUP_FOR_INFO_MSG):
                        currentFeed.popupForInfoMsg = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_POPUP_FOR_IMPORTANT_MSG):
                        currentFeed.popupForImportantMsg = self.toBoolean(strContent)
                    if(node.nodeName == CERNFeedConfiguration.N_POPUP_FOR_ALERT_MSG):
                        currentFeed.popupForAlertMsg = self.toBoolean(strContent)
                    
                # Walk the child nodes
                if(isRoot):
                    self.__readFeedConfig(node, False, tmpFeed)
                # add current feed to feeds list
                if(isRoot):
                    self.feeds.append(tmpFeed)
    
    # save feeds in feed config file                
    def __saveFeedConfig(self):
        # Create the minidom document
        doc = Document()
        
        # Create the <docRoot> base element
        docRoot = doc.createElement(CERNFeedConfiguration.N_DOCUMENT_ROOT)
        doc.appendChild(docRoot)
        
        # create feed elements
        for feed in self.feeds:
            feedElement = doc.createElement(CERNFeedConfiguration.N_ROOT)
            docRoot.appendChild(feedElement)
            
            urlElement = doc.createElement(CERNFeedConfiguration.N_URL)
            feedElement.appendChild(urlElement)
            urlText = doc.createTextNode(feed.url)
            urlElement.appendChild(urlText)
            
            nameElement = doc.createElement(CERNFeedConfiguration.N_NAME)
            feedElement.appendChild(nameElement)
            nameText = doc.createTextNode(feed.name)
            nameElement.appendChild(nameText)
            
            mandatoryElement = doc.createElement(CERNFeedConfiguration.N_MANDATORY)
            feedElement.appendChild(mandatoryElement)
            mandatoryText = doc.createTextNode(self.__convertBoolToStr(feed.mandatory))
            mandatoryElement.appendChild(mandatoryText)
            
            CERNFeedElement = doc.createElement(CERNFeedConfiguration.N_CERN_FEED)
            feedElement.appendChild(CERNFeedElement)
            CERNFeedText = doc.createTextNode(self.__convertBoolToStr(feed.CERNFeed))
            CERNFeedElement.appendChild(CERNFeedText)
            
            disabledElement = doc.createElement(CERNFeedConfiguration.N_DISABLED)
            feedElement.appendChild(disabledElement)
            disabled = doc.createTextNode(self.__convertBoolToStr(feed.disabled))
            disabledElement.appendChild(disabled)
            
            balloonForInfoMsgElement = doc.createElement(CERNFeedConfiguration.N_BALOON_FOR_INFO_MSG)
            feedElement.appendChild(balloonForInfoMsgElement)
            balloonForInfoMsgText = doc.createTextNode(self.__convertBoolToStr(feed.balloonForInfoMsg))
            balloonForInfoMsgElement.appendChild(balloonForInfoMsgText)
            
            balloonForImportantMsgElement = doc.createElement(CERNFeedConfiguration.N_BALLOON_FOR_IMPORTANT_MSG)
            feedElement.appendChild(balloonForImportantMsgElement)
            balloonForImportantMsgText = doc.createTextNode(self.__convertBoolToStr(feed.balloonForImportantMsg))
            balloonForImportantMsgElement.appendChild(balloonForImportantMsgText)
            
            balloonForAlertMsgElement = doc.createElement(CERNFeedConfiguration.N_BALLOON_FOR_ALERT_MSG)
            feedElement.appendChild(balloonForAlertMsgElement)
            balloonForAlertMsgText = doc.createTextNode(self.__convertBoolToStr(feed.balloonForAlertMsg))
            balloonForAlertMsgElement.appendChild(balloonForAlertMsgText)
            
            popupForInfoMsgElement = doc.createElement(CERNFeedConfiguration.N_POPUP_FOR_INFO_MSG)
            feedElement.appendChild(popupForInfoMsgElement)
            popupForInfoMsgText = doc.createTextNode(self.__convertBoolToStr(feed.popupForInfoMsg))
            popupForInfoMsgElement.appendChild(popupForInfoMsgText)
            
            popupForImportantMsgElement = doc.createElement(CERNFeedConfiguration.N_POPUP_FOR_IMPORTANT_MSG)
            feedElement.appendChild(popupForImportantMsgElement)
            popupForImportantMsgText = doc.createTextNode(self.__convertBoolToStr(feed.popupForImportantMsg))
            popupForImportantMsgElement.appendChild(popupForImportantMsgText)
            
            popupForAlertMsgElement = doc.createElement(CERNFeedConfiguration.N_POPUP_FOR_ALERT_MSG)
            feedElement.appendChild(popupForAlertMsgElement)
            popupForAlertMsgText = doc.createTextNode(self.__convertBoolToStr(feed.popupForAlertMsg))
            popupForAlertMsgElement.appendChild(popupForAlertMsgText)
        
            xmlString = doc.toprettyxml(indent="  ")
            feedsFile = self.loadFromUsersHomeDir(FEEDS_FILE)
            
            self.getLogger().debug("Saving " + feedsFile)
            
            output = open(feedsFile, 'w')
            output.write(xmlString)
            output.close()
            
            self.getLogger().debug(feedsFile + " successfully saved.")
    
    # convert string to boolean        
    def toBoolean(self, s):
        if s is True or s is False:
            return s
        s = str(s).strip().lower()
        return not s in ['false', 'f', 'n', '0', '']
    
    # convert boolean to string
    def __convertBoolToStr(self, b):
        if(b):
            return "True"
        else:
            return "False"
    
    # save application cinfig file    
    def saveConfig(self):
        self.__saveFeedConfig()
        
        self.getLogger().debug("Saving " + self.loadFromUsersHomeDir(PROPERTIES_FILE))
        # Save the properties config
        p = Properties()
        p.load(open(self.loadFromUsersHomeDir(PROPERTIES_FILE)))
        p[CERNConfiguration.P_USERNAME] = self.username
        p[CERNConfiguration.P_COMPUTERNAME] = self.computername
        p.store(open(self.loadFromUsersHomeDir(PROPERTIES_FILE), 'w'))
        
        self.getLogger().debug(self.loadFromUsersHomeDir(PROPERTIES_FILE) + " successfully saved.")

# Encapsulation class for RSS configuration XML
class CERNFeedConfiguration:
    
    # Node names
    N_DOCUMENT_ROOT = "feeds"
    N_ROOT = "feed"
    N_NAME = "name"
    N_URL = "url"
    N_MANDATORY = "mandatory"
    N_CERN_FEED = "CERNFeed"
    N_DISABLED = "disabled"
    
    N_BALOON_FOR_INFO_MSG = "balloonForInfoMsg"
    N_BALLOON_FOR_IMPORTANT_MSG = "balloonForImportantMsg"
    N_BALLOON_FOR_ALERT_MSG = "balloonForAlertMsg"
    N_POPUP_FOR_INFO_MSG = "popupForInfoMsg"
    N_POPUP_FOR_IMPORTANT_MSG = "popupForImportantMsg"
    N_POPUP_FOR_ALERT_MSG = "popupForAlertMsg"
    
    # Alerter does not save this info
    N_AUTH_REQUIRED = "authRequired"
    N_AUTH_FEED_USER = "authFeedUser"
    N_AUTH_FEED_PASS = "authFeedPass"
        
    def __init__(self):
        
        # feed URL
        self.url = None
        # feed name
        self.name = None
        # default values
        # mandatory feed (cannot be deleted)
        self.mandatory = False
        # CERN feed (append username and computername)
        self.CERNFeed = False
        
        # disable feed
        self.disabled = False
        
        # Alerter does not save auth details for feeds 
        self.authRequired = False
        self.authFeedUser = ""
        self.authFeedPass = ""
        
        # Following configuration options are not used yet.
        
        # show balloons for Information messages
        self.balloonForInfoMsg = True
        # show balloons for Important messages
        self.balloonForImportantMsg = True
        # show balloons for Alert messages
        self.balloonForAlertMsg = False
        # show popups for Information messages
        self.popupForInfoMsg = False
        # show popups for Important messages
        self.popupForImportantMsg = False
        # show popups for Alert messages
        self.popupForAlertMsg = True
        
    # print class
    def __str__(self):
        result = ""
        result += "url: " + self.url + "\n"
        result += "name: " + self.name + "\n"
        result += "mandatory: " + self.__convertBoolToStr(self.mandatory) + "\n"
        result += "CERNFeed: " + self.__convertBoolToStr(self.CERNFeed) + "\n"
        result += "balloonForInfoMsg: " + self.__convertBoolToStr(self.balloonForInfoMsg) + "\n"
        result += "balloonForImportantMsg: " + self.__convertBoolToStr(self.balloonForImportantMsg) + "\n"
        result += "balloonForAlertMsg: " + self.__convertBoolToStr(self.balloonForAlertMsg) + "\n"
        result += "popupForInfoMsg: " + self.__convertBoolToStr(self.popupForInfoMsg) + "\n"
        result += "popupForImportantMsg: " + self.__convertBoolToStr(self.popupForImportantMsg) + "\n"
        result += "popupForAlertMsg: " + self.__convertBoolToStr(self.popupForAlertMsg) + "\n"
        result += "authRequired: " + self.__convertBoolToStr(self.authRequired) + "\n"
        result += "authFeedUser: " + self.authFeedUser + "\n"
        result += "disabled: " + self.__convertBoolToStr(self.disabled) + "\n"
        return result
    
    # convert boolean to string    
    def __convertBoolToStr(self, b):
        if(b):
            return "True"
        else:
            return "False"
            
# Class responsible for feed reading          
class CERNFeedReader:

    def __init__(self):
        self._lock = threading.RLock()
    
    # check if feed is valid. Throw exception if it is not valid.
    def isFeedValid(self, url, feed=None):
        # do not load feed twice if we already have it
        if (feed == None):
            feed = feedparser.parse(url)
        if (feed.bozo == 1):
            CERNConfiguration.getInstance().getLogger().warn("Cannot read feed " + url)
#            CERNConfiguration.getInstance().getLogger().warn(feed.bozo_exception)
            # feed needs auth 
            if ((hasattr(feed, 'status') and feed.status== 401) or 
                (isinstance(feed.bozo_exception, IOError) and hasattr(feed.bozo_exception, 'code') 
                 and feed.bozo_exception.code == 401) or isinstance(feed.bozo_exception, ValueError) ):
                raise CERNNotifierFeedAuthException("Feed requires authentication. URL: " + url)
            if (isinstance(feed.bozo_exception, SAXParseException)):
                raise CERNNotifierFeedException("Cannot parse feed " + url)
            if (isinstance(feed.bozo_exception, urllib2.URLError)):
                raise CERNNotifierFeedException("Invalid feed URL: " + url)
            # unknown exception
            raise CERNNotifierFeedException("Unknown exception while reading feed " + url)
        else:
            CERNConfiguration.getInstance().getLogger().debug("Feed " + url + " parsed successfully")
    
    # get new feeds
    def getNewFeeds(self, _feed):
#        self._lock.acquire()
        result = []
        # edit CERN feeds and add username and hostname
        url = ""
        if(_feed.CERNFeed):
            utils = CERNUtils()
            url = utils.buildCERNFeedURL(_feed.url.strip())
        else:
            url = _feed.url.strip()
                            
        try:
            CERNConfiguration.getInstance().getLogger().debug("Parsing feed " + url.strip())
            
            if (_feed.authRequired) :
                # Keep in mind that feedparser lib is modified for the following code can work
                mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
                mgr.add_password(None, url, _feed.authFeedUser, _feed.authFeedPass)        
                feed = feedparser.parse(url, passwordManager=mgr)
# Old and unsecure way.  
#                modifiedURL = ""
#                authString = _feed.authFeedUser + ":" + _feed.authFeedPass + "@"
#                if (url.startswith("http://") or url.startswith("https://")) :
#                    modifiedURL = url.replace("://", "://" + authString)
#                else :
#                    # if the URL does not start with http/https we add the credentials first
#                    modifiedURL = authString + url
#                
#                feed = feedparser.parse(modifiedURL)
#                CERNConfiguration.getInstance().getLogger().debug("Password protected URL (password masked): "
#                                                                  + modifiedURL.replace(_feed.authFeedPass, "[password]"))
            else :
                # parse the feed
                feed = feedparser.parse(url)
            
            # check if feed is valid
            self.isFeedValid(url, feed)
            
            # iterate through all feed entries
            for i in range(len(feed['entries'])):
                entry = RssData()
    
                entry.title = feed['entries'][i]['title']
                entry.link = feed['entries'][i]['link']
                
                try:
                    tmpBody = feed['entries'][i]['summary']
                    # escape <div> html tag
                    if (tmpBody.startswith("<div") and tmpBody.endswith("</div>")):
                        tmpBody = tmpBody[tmpBody.find(">") + 1:tmpBody.rfind("</")]
                    entry.body = tmpBody
                except(KeyError):
                    pass
    #            if (entry.body == None or len(entry.body) == 0):
    #                print("ala bala")
                
                try:
                    entry.author = feed['entries'][i]['author']
                except(KeyError):
                    pass
                try:
                    entry.publicationDate = feed['entries'][i]['updated']
                except(KeyError):
                    pass
                # parse CERN specific feed info
                try:
                    entry.severity = feed['entries'][i]['nicealert_severity']
                except(KeyError):
                    pass
    #                print("nicealert_severity are not defined")
                try:
                    entry.msgType = feed['entries'][i]['nicealert_messagetype']
                except(KeyError):
                    pass
    #                print("nicealert_messagetype are not defined")
                try:
                    entry.validFromDate = feed['entries'][i]['nicealert_validfrom']
                except(KeyError):
                    pass
    #                print("nicealert_validfrom are not defined")
                try:
                    entry.expirationDate = feed['entries'][i]['nicealert_expirationdate']
                except(KeyError):
                    pass
    #                print("nicealert_expirationdate are not defined")
                try:
                    entry.enclosures = feed['entries'][i]['enclosures']
                except(KeyError):
                    pass
    #                print("enclosures are not defined")
                try:
                    entry.department = feed['entries'][i]['nicealert_department']
                except(KeyError):
                    pass
    #                print("nicealert_department is not defined")
                try:
                    entry.building = feed['entries'][i]['nicealert_building']
                except(KeyError):
                    pass
    #                print("nicealert_building is not defined")
                try:
                    entry.os = feed['entries'][i]['nicealert_operatingsystem']
                except(KeyError):
                    pass
    #                print("nicealert_operatingsystem is not defined")
                try:
                    entry.mailServer = feed['entries'][i]['nicealert_mailserver']
                except(KeyError):
                    pass
    #                print("nicealert_mailserver is not defined")
                try:
                    entry.computerName = feed['entries'][i]['nicealert_computername']
                except(KeyError):
                    pass
    #                print("nicealert_computername is not defined")      
                
                result.append(entry)
                CERNConfiguration.getInstance().getLogger().debug("Added feed " + entry.title + " to result set. Feed No " + str(i) + " out of " + str(len(feed['entries'])))
        finally:
            pass
#            self._lock.release()
                
        return result

# Encapsulate RSS data
class RssData:
    SEVERITY_ALERT = "ALERT"
    SEVERITY_IMPORTANT_MESSAGE = "IMPORTANT MESSAGE"
    SEVERITY_INFORMATION = "INFORMATION"
    
    title = 'None'
    body = 'None'
    author = 'None'
    publicationDate = 'None'
    link = '#'
    
    enclosures = None
    severity = 'None'
    
    msgType = 'None'
    validFromDate = 'None'
    expirationDate = 'None'
    
    department = 'None'
    building = 'None'
    mailServer = 'None'
    os = 'None'
    computerName = 'None'
    
    # print class
    def __str__(self):
        result = ""
        result += "Title: " + self.title + "\n"
        result += "Body: " + self.body + "\n"
        result += "Author: " + self.author + "\n"
        result += "Publication date: " + self.publicationDate + "\n"
        result += "Link: " + self.link + "\n"
        
        if(self.enclosures != None):
            enclosuresStr = ""
            for enclosure in self.enclosures:
                enclosuresStr += enclosure.get('href') + " "
            result += "Enclosures: " + enclosuresStr + "\n" 
        else:
            result += "Enclosures: None \n"
        
        result += "Severity: " + self.severity + "\n"
        
        result += "Message type: " + self.msgType + "\n"
        result += "Valid from: " + self.validFromDate + "\n"
        result += "Valid to: " + self.expirationDate + "\n"
        
        result += "Department: " + self.department + "\n"
        result += "Building: " + self.building + "\n"
        result += "Mail Server: " + self.mailServer + "\n"
        result += "OS: " + self.os + "\n"
        result += "Computer Name: " + self.computerName + "\n"
        
        return result
    
#    def __cmp__(self, other):
#        return self.link == other.link
    
    # 2 feeds are equal if their link is the same
    def __eq__(self, other):
        return self.link == other.link


def run():
    pass
    #os.system('python CERNNotifierUI.py')
    

def info():
    print """ TODO: Print some info""" 

if __name__ == '__main__':
    run()
